import re

reddit_links_ignore = re.compile(
    "|".join(
        r"""youtube.com/user/
youtube.com/c/
youtube.com/channel/
youtube.com/results""".splitlines()
    )
)

yt_recoverable_errors = re.compile(
    "|".join(
        r"""due to geo restriction
geo-restricted
geolocation
your country
No such file or directory
fragment 1 not found
HTTP Error 429
Too Many Requests
Premieres in
read operation timed out
Internal Server Error
Internal error encountered
Playlists that require authentication may not extract correctly without a successful webpage download
Main webpage is locked behind the login page
You need to log in to access this content
This video is only available for registered users
Could not send HEAD request
Unable to download JSON metadata
Failed to parse JSON Expecting
expected string or bytes-like object
Connection refused
giving up after.*retries
Failed to download MPD manifest:\$
not currently available\$
copyright claim""".splitlines()
    )
)


yt_meaningless_errors = re.compile(
    "|".join(
        r"""hidden
timed out
Timeout
Timed
Connection reset
ConnectionReset
Unauthorized
Forbidden
Traceback
KeyboardInterrupt
Fatal Python error
list index out of range
Extract.* cookies
File .*, line .*, in
Requested format is not available.
fragment_filename_sanitized
no suitable InfoExtractor for URL
No such file or directory
: Downloading webpage\$
: Extracting information\$
: Requesting header\$
Downloading .* metadata\$
Downloading .* information\$
Downloading .* manifest\$
Determining source extension\$
Downloading jwt token\$
^\[info\]
^\[redirect\]
^\[Merger\]
^\[dashsegments\]
Finished downloading playlist
The last 30x error message was:
^Found\$
NoneType
Creating a generic title instead
The channel is not currently live
clips are not currently supported.
Join this channel to get access to members-only content
Confirm you are on the latest version using
referenced before assignment
field is missing or empty
list index out of range
Interrupted by user
unable to open for writing:
You might want to use a VPN or a proxy server
maximum recursion depth exceeded
object does not support item assignment
encodings are not supported
object has no attribute
merged
Compressed file ended before the end-of-stream marker was reached
Falling back on generic
Some formats are possibly damaged
matching opening tag for closing p tag not found
the JSON object must be str, bytes or bytearray, not dict
The read operation timed out
Unable to recognize playlist.
Premieres in""".splitlines()
    )
)


yt_unrecoverable_errors = re.compile(
    "|".join(
        r"""repetitive or misleading metadata
It is not available
has already been recorded in the archive
ideo.*is private
Private video
This playlist is private
Unable to recognize tab page
'bytes' object has no attribute 'encode'
'NoneType' object has no attribute 'get'
list index out of range
Unable to extract cnn url
PornHd.*Unable to extract error message
members-only content
You don't have permission to access this video.
Video is unavailable pending review
Video has been flagged for verification
This video has been disabled\$
The uploader has not made this video available.\$
This video is DRM protected
This video is protected by a password
This video requires payment to watch.\$
Unable to download webpage
dashboard-only post
This video is only available to Music Premium members
The policy key provided does not permit this account or video
live stream recording
The channel is not currently live
This live event will begin in a few moments
nudity or sexual content
policy on harassment and bullying
stream.* is offline\$
: Video unavailable\$
 does not exist.\$
 has been removed\$
Premieres in.*hours\$
This clip is no longer available\$
No media found\$
No sources found for video
caused by KeyError\('title'
\[youtube:truncated_url\]
: This video is not available.\$
Video unavailable. This video is not available\$
Resource temporarily unavailable
This video is unavailable
Unsupported URL
URL could be a direct video link
not a valid URL
not a video\$
The page doesn't contain any tracks
removed by the uploader
blocked it on copyright grounds\$
uploader has closed their.*account
account has been terminated because we received multiple
User has been suspended
policy on violent or graphic content\$
The channel does not have a .* tab\$
policy on spam, deceptive practices, and scams\$
This video does not exist, or has been deleted.
Community Guidelines
Terms of Service
This channel does not exist
account associated with this video
This video doesn't exist.\$
Track not found\$
Not found.\$
Can't find object media for
o video formats found
o video in
o video on
certificate is not valid
CERTIFICATE_VERIFY_FAILED
HTTP Error 403: Forbidden\$
code -404
HTTP Error 404
HTTPError 404
HTTP Error 410
HTTPError 410""".splitlines()
    )
)



import sys
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup

from xklb.utils import log


def get_page_links(url):
    soup = BeautifulSoup(requests.get(url).content, "html.parser")
    film_list = set()

    for a in soup.findAll("a", attrs={"href": True}):
        href = a["href"].strip()
        if all([len(href) > 1, href[0] != "#", "javascript:" not in href, "mailto:" not in href, "tel:" not in href]):
            if "/film/" in href:
                up = urlparse(url)
                film_list.add(up.scheme + "://" + up.netloc + href)
            else:
                log.debug(href)

    for el in film_list:
        print(el)


def nfb_films() -> None:
    for l in sys.stdin:
        l = l.rstrip("\n")
        if l in ["", '""', "\n"]:
            continue

        get_page_links(l)


if __name__ == "__main__":
    # echo $directors | python scripts/mining/nfb.ca.py | tee -a ~/.jobs/todo/71_Mealtime_Videos
    nfb_films()



from scripts.deduplicate_music import deduplicate_music
from scripts.invalid_filenames import rename_invalid_files
from scripts.large_folders import large_folders
from scripts.merge_online_local import merge_online_local
from scripts.mining.nfb_ca import nfb_films
from scripts.mining.pushshift import pushshift_extract
from scripts.mining.reddit_self import parse_reddit_selftext
from scripts.optimize_db import optimize_db
from scripts.streaming_tab_loader import streaming_tab_loader



import argparse, sys
from pathlib import Path

import gallery_dl as gdl

from xklb import consts, db, utils
from xklb.consts import SC, DBType
from xklb.utils import log


def parse_args(action, usage) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="library " + action, usage=usage)

    parser.add_argument(
        "--dl-config",
        "-dl-config",
        nargs=1,
        action=utils.argparse_dict,
        default={},
        metavar="KEY=VALUE",
        help="Add key/value pairs to override or extend default downloader configuration",
    )
    parser.add_argument("--safe", "-safe", action="store_true", help="Skip generic URLs")
    parser.add_argument("--no-sanitize", "-s", action="store_true", help="Don't sanitize some common URL parameters")
    parser.add_argument("--extra", "-extra", action="store_true", help="Get full metadata (takes a lot longer)")
    parser.add_argument("--extra-media-data", default={})
    parser.add_argument("--extra-playlist-data", default={})
    parser.add_argument("--ignore-errors", "--ignoreerrors", "-i", action="store_true", help=argparse.SUPPRESS)

    if action in (SC.galyadd, SC.galyupdate):
        parser.add_argument("--category", "-c", help=argparse.SUPPRESS)

    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)

    parser.add_argument("database", nargs="?", default="video.db")
    if action in (SC.galyadd):
        parser.add_argument("playlists", nargs="+", help=argparse.SUPPRESS)

    args = parser.parse_args()
    args.action = action

    if args.db:
        args.database = args.db
    if action in (SC.galyadd):
        Path(args.database).touch()
    args.db = db.connect(args)

    if hasattr(args, "no_sanitize") and hasattr(args, "playlists") and not args.no_sanitize:
        args.playlists = [consts.sanitize_url(args, p) for p in args.playlists]

    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def galy_add(args=None) -> None:
    if args:
        sys.argv[1:] = args

    args = parse_args(
        SC.galyadd,
        usage=r"""library galyadd [--audio | --video] -c CATEGORY [database] playlists ...

    Create a dl database / add links to an existing database

        library galyadd -c Educational dl.db https://www.youdl.com/c/BranchEducation/videos

    If you include more than one URL, you must specify the database

        library galyadd 71_Mealtime_Videos dl.db (cat ~/.jobs/todo/71_Mealtime_Videos)

    Files will be saved to <lb download prefix>/<lb galyadd category>/

        For example:
        library galyadd Cool ...
        library download D:\'My Documents'\ ...
        Media will be downloaded to 'D:\My Documents\Cool\'

    Fetch extra metadata:

        By default galyadd will quickly add media at the expense of less metadata.
        If you plan on using `library download` then it doesn't make sense to use `--extra`.
        Downloading will add the extra metadata automatically to the database.
        You can always fetch more metadata later via galyupdate:
        library galyupdate tw.db --extra
    """,
    )
    for path in args.playlists:

        gdl.config.load()  # load default config files
        job = gdl.job.DataJob(path)
        job.run()
        urls = job.data
        raise
        # TODO: save gallery-dl data to the database
        # gdl.config.set(("extractor",), "base-directory", "/tmp/")
        # gdl.job.DownloadJob(path)



import os, sqlite3
from typing import Any, Iterable, List, Optional, Union

import sqlite_utils

from xklb.utils import log


class DB(sqlite_utils.Database):
    def pop(self, sql: str, params: Optional[Union[Iterable, dict]] = None, ignore_errors=None) -> Optional[Any]:
        if ignore_errors is None:
            ignore_errors = ["no such table"]
        try:
            curs = self.execute(sql, params)
            data = curs.fetchone()
        except sqlite3.OperationalError as exc:
            if any([e in str(exc) for e in ignore_errors]):
                return None
            raise
        if data is None or len(data) == 0:
            return None
        return data[0]

    def pop_dict(self, sql: str, params: Optional[Union[Iterable, dict]] = None, ignore_errors=None) -> Optional[Any]:
        if ignore_errors is None:
            ignore_errors = ["no such table"]
        try:
            dg = self.query(sql, params)
            d = next(dg, None)
        except sqlite3.OperationalError as exc:
            if any([e in str(exc) for e in ignore_errors]):
                return None
            raise exc
        return d


def tracer(sql, params) -> None:
    print("SQL: {} - params: {}".format(sql, params))


def connect(args, conn=None, **kwargs) -> sqlite_utils.Database:
    if not os.path.exists(args.database) and ":memory:" not in args.database:
        log.error(f"Database file '{args.database}' does not exist. Create one with lb fsadd, tubeadd, or tabsadd.")
        raise SystemExit(1)

    db = DB(conn or args.database, tracer=tracer if args.verbose >= 2 else None, **kwargs)  # type: ignore
    with db.conn:
        db.conn.execute("PRAGMA main.cache_size = 8000")
    return db


def optimize(args) -> None:
    print("\nOptimizing database")
    db: sqlite_utils.Database = args.db
    tables = db.table_names()
    db.enable_wal()

    config = {
        "media": {
            "search_columns": ["path", "title", "tags", "mood", "genre", "description", "artist", "album"],
            "column_order": ["path", "webpath", "id", "ie_key", "playlist_path"],
            "ignore_columns": ["id"],
        },
        "reddit_posts": {
            "search_columns": ["title", "selftext_html"],
        },
        "reddit_comments": {
            "search_columns": ["body"],
        },
    }

    for table in ["media", "playlists", "reddit_posts", "reddit_comments"]:
        if table in tables:
            table_columns = db[table].columns_dict
            table_config = config.get(table) or {}
            ignore_columns = table_config.get("ignore_columns") or []
            search_columns = table_config.get("search_columns") or []

            fts_columns = [c for c in search_columns if c in table_columns]
            int_columns = [k for k, v in table_columns.items() if v == int and k not in search_columns + ignore_columns]
            str_columns = [k for k, v in table_columns.items() if v == str and k not in search_columns + ignore_columns]

            optimized_column_order = [*int_columns, *(table_config.get("column_order") or [])]
            current_order = zip(table_columns, optimized_column_order)
            was_transformed = False
            if not all([x == y for x, y in current_order]):
                db[table].transform(column_order=optimized_column_order)  # type: ignore
                was_transformed = True

            for column in int_columns + str_columns:
                db[table].create_index([column], if_not_exists=True, analyze=True)  # type: ignore

            if any(fts_columns) and (db[table].detect_fts() is None or was_transformed):  # type: ignore
                db[table].enable_fts(fts_columns, create_triggers=True, replace=True)

            with db.conn:
                db[table].optimize()  # type: ignore

    db.vacuum()
    db.analyze()


def fts_quote(query: List[str]) -> List[str]:
    fts_words = [" NOT ", " AND ", " OR ", "*", ":", "NEAR("]
    return [s if any([r in s for r in fts_words]) else '"' + s + '"' for s in query]


def fts_search(args, bindings) -> str:
    bindings["query"] = " AND ".join(fts_quote(args.include))
    if args.exclude:
        bindings["query"] += " NOT " + " NOT ".join(fts_quote(args.exclude))
    table = "(" + args.db["media"].search_sql() + ")"  #  include_rank=True
    return table


def gen_include_excludes(cols_available):
    searchable_columns = [
        "path",
        "title",
        "mood",
        "genre",
        "year",
        "bpm",
        "key",
        "time",
        "decade",
        "categories",
        "city",
        "country",
        "description",
        "album",
        "artist",
        "tags",
        "playlist_path",
    ]

    valid_cols = [f"media.{c}" for c in searchable_columns if c in cols_available]

    include_string = "and (" + " like :include{} OR ".join(valid_cols) + " like :include{} )"
    exclude_string = "and (" + " not like :exclude{} AND ".join(valid_cols) + " not like :exclude{} )"

    return include_string, exclude_string


def get_playlists(args, cols="path, dl_config", constrain=False, sql_filters=None) -> List[dict]:
    columns = args.db["playlists"].columns_dict
    if sql_filters is None:
        sql_filters = []
    if "time_deleted" in columns:
        sql_filters.append("AND time_deleted=0")
    if constrain:
        if args.category:
            sql_filters.append(f"AND category='{args.category}'")

    try:
        known_playlists = list(
            args.db.query(f"select {cols} from playlists where 1=1 {' '.join(sql_filters)} order by random()")
        )
    except sqlite3.OperationalError:
        known_playlists = []
    return known_playlists


def get_playlists_join(args):
    media_columns = args.db["media"].columns_dict
    if "ie_key" in media_columns:
        join = "(p.ie_key = media.ie_key = 'Local' and media.path like p.path || '%' ) "
        if "playlist_path" in media_columns:
            join += "or (p.ie_key = media.ie_key and media.ie_key != 'Local' and p.path = media.playlist_path)"
    else:
        join = "p.path = media.playlist_path"

    return join



import argparse
from time import sleep
from typing import Dict, List, Tuple

from xklb import db, utils
from xklb.consts import SC
from xklb.player import generic_player, mark_media_watched, override_sort, printer
from xklb.tabs_extract import Frequency
from xklb.utils import cmd, flatten, log

tabs_include_string = (
    lambda x: f"""and (
    path like :include{x}
    OR category like :include{x}
    OR frequency like :include{x}
)"""
)

tabs_exclude_string = (
    lambda x: f"""and (
    path not like :exclude{x}
    AND category not like :exclude{x}
    AND frequency not like :exclude{x}
)"""
)


def construct_tabs_query(args) -> Tuple[str, dict]:
    cf = []
    bindings = {}

    cf.extend([" and " + w for w in args.where])

    for idx, inc in enumerate(args.include):
        cf.append(tabs_include_string(idx))
        bindings[f"include{idx}"] = "%" + inc.replace(" ", "%").replace("%%", " ") + "%"
    for idx, exc in enumerate(args.exclude):
        cf.append(tabs_exclude_string(idx))
        bindings[f"exclude{idx}"] = "%" + exc.replace(" ", "%").replace("%%", " ") + "%"

    args.sql_filter = " ".join(cf)

    LIMIT = "LIMIT " + str(args.limit) if args.limit else ""
    OFFSET = f"OFFSET {args.skip}" if args.skip else ""

    query = f"""SELECT path
        , frequency
        , CASE
            WHEN frequency = 'daily' THEN cast(STRFTIME('%s', datetime( time_played, 'unixepoch', '+1 Day', '-3 hours' )) as int)
            WHEN frequency = 'weekly' THEN cast(STRFTIME('%s', datetime( time_played, 'unixepoch', '+7 Days', '-3 hours' )) as int)
            WHEN frequency = 'monthly' THEN cast(STRFTIME('%s', datetime( time_played, 'unixepoch', '+1 Month', '-3 hours' )) as int)
            WHEN frequency = 'quarterly' THEN cast(STRFTIME('%s', datetime( time_played, 'unixepoch', '+3 Months', '-3 hours' )) as int)
            WHEN frequency = 'yearly' THEN cast(STRFTIME('%s', datetime( time_played, 'unixepoch', '+1 Year', '-3 hours' )) as int)
        END time_valid
        {', ' + ', '.join(args.cols) if args.cols else ''}
    FROM media
    WHERE 1=1
        {args.sql_filter}
        {"and time_valid < cast(STRFTIME('%s', datetime()) as int)" if not args.print else ''}
    ORDER BY 1=1
        {', ' + args.sort if args.sort else ''}
        {', time_played, time_valid, path' if args.print else ''}
        , play_count
        , frequency = 'daily' desc
        , frequency = 'weekly' desc
        , frequency = 'monthly' desc
        , frequency = 'quarterly' desc
        , frequency = 'yearly' desc
        , ROW_NUMBER() OVER ( PARTITION BY
            play_count
            , frequency
            , hostname
            , category
        ) -- prefer to spread hostname, category over time
        , random()
    {LIMIT} {OFFSET}
    """

    return query, bindings


def play(args, m: Dict) -> None:
    media_file = m["path"]

    cmd(*generic_player(args), media_file, strict=False)
    mark_media_watched(args, media_file)


def frequency_filter(args, media: List[Dict]) -> List[dict]:
    mapper = {
        Frequency.Daily.value: 1,
        Frequency.Weekly.value: 7,
        Frequency.Monthly.value: 30,
        Frequency.Quarterly.value: 91,
        Frequency.Yearly.value: 365,
    }
    counts = args.db.execute("select frequency, count(*) from media group by 1").fetchall()
    filtered_media = []
    for freq, freq_count in counts:
        num_days = mapper.get(freq, 365)
        num_tabs = max(1, freq_count // num_days)
        log.debug(f"freq_count {freq_count} // num_days {num_days} = num_tabs {num_tabs}")

        t = []
        for m in media:
            if m["frequency"] == freq:
                t.append(m)

        filtered_media.extend(t[:num_tabs])

    return filtered_media


def process_tabs_actions(args) -> None:
    query, bindings = construct_tabs_query(args)

    if args.print:
        return printer(args, query, bindings)

    media = list(args.db.query(query, bindings))
    if not media:
        utils.no_media_found()

    media = frequency_filter(args, media)

    for m in media:
        play(args, m)
        if len(media) >= 9:
            sleep(0.3)


def parse_args(action, default_db) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="library tabs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage="""library tabs [database] [optional args]

    Tabs is meant to run **once per day**. Here is how you would configure it with `crontab`:

        45 9 * * * DISPLAY=:0 library tabs /home/my/tabs.db

    If things aren't working you can use `at` to simulate a similar environment as `cron`

        echo 'fish -c "export DISPLAY=:0 && library tabs /full/path/to/tabs.db"' | at NOW

    You can also invoke tabs manually:

        library tabs -L 1  # open one tab

    Print URLs

        lb-dev tabs -w "frequency='yearly'" -p
        ╒════════════════════════════════════════════════════════════════╤═════════════╤══════════════╕
        │ path                                                           │ frequency   │ time_valid   │
        ╞════════════════════════════════════════════════════════════════╪═════════════╪══════════════╡
        │ https://old.reddit.com/r/Autonomia/top/?sort=top&t=year        │ yearly      │ Dec 31 1970  │
        ├────────────────────────────────────────────────────────────────┼─────────────┼──────────────┤
        │ https://old.reddit.com/r/Cyberpunk/top/?sort=top&t=year        │ yearly      │ Dec 31 1970  │
        ├────────────────────────────────────────────────────────────────┼─────────────┼──────────────┤
        │ https://old.reddit.com/r/ExperiencedDevs/top/?sort=top&t=year  │ yearly      │ Dec 31 1970  │

        ...

        ╘════════════════════════════════════════════════════════════════╧═════════════╧══════════════╛

    View how many yearly tabs you have:

        library tabs -w "frequency='yearly'" -p a
        ╒═══════════╤═════════╕
        │ path      │   count │
        ╞═══════════╪═════════╡
        │ Aggregate │     134 │
        ╘═══════════╧═════════╛

    Delete URLs

        library tb -p -s cyber
        ╒═══════════════════════════════════════╤═════════════╤══════════════╕
        │ path                                  │ frequency   │ time_valid   │
        ╞═══════════════════════════════════════╪═════════════╪══════════════╡
        │ https://old.reddit.com/r/cyberDeck/to │ yearly      │ Dec 31 1970  │
        │ p/?sort=top&t=year                    │             │              │
        ├───────────────────────────────────────┼─────────────┼──────────────┤
        │ https://old.reddit.com/r/Cyberpunk/to │ yearly      │ Aug 29 2023  │
        │ p/?sort=top&t=year                    │             │              │
        ├───────────────────────────────────────┼─────────────┼──────────────┤
        │ https://www.reddit.com/r/cyberDeck/   │ yearly      │ Sep 05 2023  │
        ╘═══════════════════════════════════════╧═════════════╧══════════════╛
        library tb -p -w "path='https://www.reddit.com/r/cyberDeck/'" --delete
        Removed 1 metadata records
        library tb -p -s cyber
        ╒═══════════════════════════════════════╤═════════════╤══════════════╕
        │ path                                  │ frequency   │ time_valid   │
        ╞═══════════════════════════════════════╪═════════════╪══════════════╡
        │ https://old.reddit.com/r/cyberDeck/to │ yearly      │ Dec 31 1970  │
        │ p/?sort=top&t=year                    │             │              │
        ├───────────────────────────────────────┼─────────────┼──────────────┤
        │ https://old.reddit.com/r/Cyberpunk/to │ yearly      │ Aug 29 2023  │
        │ p/?sort=top&t=year                    │             │              │
        ╘═══════════════════════════════════════╧═════════════╧══════════════╛
""",
    )

    parser.add_argument("--sort", "-u", nargs="+")
    parser.add_argument("--where", "-w", nargs="+", action="extend", default=[])
    parser.add_argument("--include", "-s", "--search", nargs="+", action="extend", default=[])
    parser.add_argument("--exclude", "-E", "-e", nargs="+", action="extend", default=[])
    parser.add_argument("--print", "-p", default=False, const="p", nargs="?")
    parser.add_argument("--delete", "--remove", "--erase", "--rm", "-rm", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--cols", "-cols", "-col", nargs="*", help="Include a non-standard column when printing")
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue")
    parser.add_argument("--skip", "-S")

    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--db", "-db")

    parser.add_argument(
        "database",
        nargs="?",
        default=default_db,
        help="Database file. If not specified a generic name will be used: audio.db, video.db, fs.db, etc",
    )
    args = parser.parse_args()
    args.action = action

    if args.db:
        args.database = args.db

    if args.sort:
        args.sort = " ".join(args.sort)
        args.sort = override_sort(args.sort)

    if args.cols:
        args.cols = list(flatten([s.split(",") for s in args.cols]))

    if args.delete:
        args.print += "d"

    if args.db:
        args.database = args.db
    args.db = db.connect(args)
    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def tabs() -> None:
    args = parse_args(SC.tabs, "tabs.db")
    process_tabs_actions(args)



import argparse, json, sys
from pathlib import Path

from xklb import db, praw_extract, utils
from xklb.utils import log


def parse_args(action, usage) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="library " + action, usage=usage)
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)
    parser.add_argument("database", nargs="?", default="pushshift.db")
    args = parser.parse_args()

    if args.db:
        args.database = args.db
    Path(args.database).touch()
    args.db = db.connect(args)

    log.info(utils.dict_filter_bool(args.__dict__))

    return args

def pushshift_extract(args=None) -> None:
    if args:
        sys.argv[1:] = args

    args = parse_args(
        "pushshift",
        usage="""library pushshift [database] < stdin

    Download data (about 600GB jsonl.zst; 6TB uncompressed)

        wget -e robots=off -r -k -A zst https://files.pushshift.io/reddit/submissions/

    Load data from files via unzstd

        unzstd --memory=2048MB --stdout RS_2005-07.zst | lb pushshift pushshift.db

    Or multiple:

        for f in files.pushshift.io/reddit/submissions/*.zst
            echo "$f"
            unzstd --memory=2048MB --stdout "$f" | lb pushshift pushshift.db
        end
    """,
    )

    for l in sys.stdin:
        l = l.rstrip("\n")
        if l in ["", '""', "\n"]:
            continue

        post_dict = json.loads(l)
        praw_extract.save_post(args, post_dict, post_dict['subreddit'], upsert=False)

    if not args.db["media"].detect_fts():
        db.optimize(args)



import argparse, sys
from datetime import datetime
from pathlib import Path
from typing import List
from urllib.parse import urlparse

from xklb import db, player, utils
from xklb.consts import Frequency, sanitize_url
from xklb.utils import argparse_enum, log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="library tabsadd",
        usage=r"""library tabsadd --frequency {daily,weekly,monthly,quarterly,yearly} --category CATEGORY [--no-sanitize] [database] paths ...

    Adding one URL:

        library tabsadd -f monthly -c travel ~/lb/tabs.db https://old.reddit.com/r/Colombia/top/?sort=top&t=month

        Depending on your shell you may need to escape the URL (add quotes)

        If you use Fish shell know that you can enable features to make pasting easier:
            set -U fish_features stderr-nocaret qmark-noglob regex-easyesc ampersand-nobg-in-token

        Also I recommend turning Ctrl+Backspace into a super-backspace for repeating similar commands with long args:
            echo 'bind \b backward-kill-bigword' >> ~/.config/fish/config.fish

    Importing from a line-delimitated file:

        library tabsadd -f yearly -c reddit ~/lb/tabs.db (cat ~/mc/yearly-subreddit.cron)

""",
    )
    parser.add_argument(
        "--frequency",
        "--freqency",
        "-f",
        default=Frequency.Monthly,
        type=Frequency,
        action=argparse_enum,
        help=argparse.SUPPRESS,
    )
    parser.add_argument("--category", "-c", help=argparse.SUPPRESS)
    parser.add_argument("--no-sanitize", "-s", action="store_true", help="Don't sanitize some common URL parameters")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)

    parser.add_argument("database", nargs="?", default="tabs.db")
    parser.add_argument("paths", nargs="+")
    args = parser.parse_args()

    if args.db:
        args.database = args.db

    Path(args.database).touch()
    args.db = db.connect(args)

    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def get_new_paths(args) -> List[str]:
    if not args.no_sanitize:
        args.paths = [sanitize_url(args, path) for path in args.paths]

    if args.category is None:
        qb = (
            "select path from media where path in (" + ",".join(["?"] * len(args.paths)) + ")",
            (*args.paths,),
        )
    else:
        qb = (
            "select path from media where category = ? and path in (" + ",".join(["?"] * len(args.paths)) + ")",
            (args.category, *args.paths),
        )

    try:
        existing = set([d["path"] for d in args.db.query(*qb)])
    except Exception:
        pass
    else:
        if existing:
            print(f"Updating frequency for {len(existing)} existing paths")
            player.mark_media_deleted(args, list(existing))

    args.paths = utils.conform([path.strip() for path in args.paths])
    return args.paths


def extract_url_metadata(args, path: str) -> dict:
    hostname = urlparse(path).hostname or ""

    return dict(
        path=path,
        hostname=hostname,
        frequency=args.frequency.value,
        category=args.category or "Uncategorized",
        time_created=int(datetime.now().timestamp()),
        time_played=0,
        play_count=0,
        time_deleted=0,
    )


def tabs_add(args=None) -> None:
    if args:
        sys.argv[1:] = args
    args = parse_args()

    tabs = [extract_url_metadata(args, path) for path in get_new_paths(args)]
    args.db["media"].insert_all(utils.list_dict_filter_bool(tabs), pk="path", alter=True, replace=True)



import argparse, math, os, sys
from multiprocessing import TimeoutError as mp_TimeoutError
from pathlib import Path
from timeit import default_timer as timer
from typing import Dict, List, Optional

from joblib import Parallel, delayed

from xklb import av, books, consts, db, utils
from xklb.consts import SC, DBType
from xklb.player import mark_media_deleted
from xklb.utils import log


def calculate_sparseness(stat) -> int:
    if stat.st_size == 0:
        sparseness = 0
    else:
        blocks_allocated = stat.st_blocks * 512
        sparseness = blocks_allocated / stat.st_size
    return sparseness


def extract_metadata(mp_args, f) -> Optional[Dict[str, int]]:
    log.debug(f)

    try:
        stat = os.stat(f)
    except Exception:
        log.error(f"[{f}] Could not read file stats (possible filesystem corruption; check dmesg)")
        return

    media = {
        "path": f,
        "play_count": 0,
        "time_played": 0,
        "size": stat.st_size,
        "time_created": int(stat.st_ctime),
        "time_modified": int(stat.st_mtime),
        "time_downloaded": consts.NOW,
        "time_deleted": 0,
        "ie_key": "Local",
    }

    if hasattr(stat, "st_blocks"):
        media = {**media, "sparseness": calculate_sparseness(stat)}

    if mp_args.profile == DBType.filesystem:
        media = {**media, "is_dir": os.path.isdir(f)}

    if mp_args.profile in (DBType.audio, DBType.video):
        return av.munge_av_tags(mp_args, media, f)

    if mp_args.profile == DBType.text:
        try:
            start = timer()
            if any([mp_args.ocr, mp_args.speech_recognition]):
                media = books.munge_book_tags_slow(media, f)
            else:
                media = books.munge_book_tags_fast(media, f)
        except mp_TimeoutError:
            log.warning(f"[{f}]: Timed out trying to read file")
            return media
        else:
            log.debug(f"[{f}]: {timer()-start}")
            return media

    return media


def extract_chunk(args, chunk_paths) -> None:
    n_jobs = -1
    if args.profile in (DBType.text, DBType.image, DBType.filesystem):
        n_jobs = consts.CPU_COUNT
    if args.verbose >= 2:
        n_jobs = 1

    mp_args = argparse.Namespace(**{k: v for k, v in args.__dict__.items() if k not in {"db"}})
    metadata = Parallel(n_jobs=n_jobs)(delayed(extract_metadata)(mp_args, p) for p in chunk_paths) or []

    if args.profile == DBType.image:
        metadata = books.extract_image_metadata_chunk(metadata, chunk_paths)

    if args.scan_subtitles:
        for p in Path(consts.SUB_TEMP_DIR).glob("*.srt"):
            p.unlink()

    media = list(filter(None, metadata))
    args.db["media"].insert_all(utils.list_dict_filter_bool(media), pk="path", alter=True, replace=True)


def find_new_files(args, path: Path) -> List[str]:
    if args.profile == DBType.audio:
        scanned_files = consts.get_media_files(path, audio=True)
    elif args.profile == DBType.video:
        scanned_files = consts.get_media_files(path)
    elif args.profile == DBType.text:
        scanned_files = consts.get_text_files(path, OCR=args.ocr, speech_recognition=args.speech_recognition)
    elif args.profile == DBType.image:
        scanned_files = consts.get_image_files(path)
    elif args.profile == DBType.filesystem:
        # thanks to these people for making rglob fast https://bugs.python.org/issue26032
        scanned_files = [str(p) for p in path.rglob("*")]
    else:
        raise Exception(f"fs_extract for profile {args.profile} not implemented")

    columns = args.db["media"].columns_dict
    scanned_set = set(scanned_files)

    try:
        existing_set = set(
            [
                d["path"]
                for d in args.db.query(
                    f"""select path from media
                where 1=1
                    and time_deleted=0
                    and path like '{path}%'
                    {'AND time_downloaded > 0' if 'time_downloaded' in columns else ''}
                """
                )
            ]
        )
    except Exception as e:
        log.debug(e)
        new_files = list(scanned_set)
    else:
        new_files = list(scanned_set - existing_set)

        deleted_files = list(existing_set - scanned_set)
        if not new_files and len(deleted_files) >= len(existing_set) and not args.force:
            return []  # if path not mounted or all files deleted
        deleted_count = mark_media_deleted(args, deleted_files)
        if deleted_count > 0:
            print(f"[{path}] Marking", deleted_count, "orphaned metadata records as deleted")

    return new_files


def _add_folder(args, folder_path: Path) -> None:
    category = args.category or folder_path.parts[-1]

    playlist = {
        "ie_key": "Local",
        "path": str(folder_path),
        "config": utils.filter_namespace(args, ["ocr", "speech_recognition", "scan_subtitles"]),
        "time_deleted": 0,
        "category": category,
        **args.extra_playlist_data,
    }
    args.db["playlists"].upsert(utils.dict_filter_bool(playlist), pk="path", alter=True)


def scan_path(args, path_str: str) -> int:
    path = Path(path_str).resolve()
    if not path.exists():
        print(f"[{path}] Path does not exist")
        if not args.force:
            return 0
    print(f"[{path}] Building file list...")

    new_files = find_new_files(args, path)

    if new_files:
        print(f"[{path}] Adding {len(new_files)} new media")
        log.debug(new_files)

        if args.profile in (DBType.text):
            batch_count = consts.CPU_COUNT
        else:
            batch_count = consts.SQLITE_PARAM_LIMIT // 100
        chunks_count = math.ceil(len(new_files) / batch_count)
        df_chunked = utils.chunks(new_files, batch_count)
        for idx, l in enumerate(df_chunked):
            percent = ((batch_count * idx) + len(l)) / len(new_files) * 100
            print(f"[{path}] Extracting metadata {percent:3.1f}% (chunk {idx + 1} of {chunks_count})")
            extract_chunk(args, l)

    _add_folder(args, path)

    return len(new_files)


def parse_args(action, usage) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="library " + action, usage=usage)

    profile = parser.add_mutually_exclusive_group()
    profile.add_argument(
        "--audio", "-A", action="store_const", dest="profile", const=DBType.audio, help="Create audio database"
    )
    profile.add_argument(
        "--filesystem",
        "-F",
        action="store_const",
        dest="profile",
        const=DBType.filesystem,
        help="Create filesystem database",
    )
    profile.add_argument(
        "--video", "-V", action="store_const", dest="profile", const=DBType.video, help="Create video database"
    )
    profile.add_argument(
        "--text", "-T", action="store_const", dest="profile", const=DBType.text, help="Create text database"
    )
    profile.add_argument(
        "--image", "-I", action="store_const", dest="profile", const=DBType.image, help="Create image database"
    )
    parser.set_defaults(profile=DBType.video)
    parser.add_argument("--category", "-c", help=argparse.SUPPRESS)

    parser.add_argument("--ocr", "--OCR", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--speech-recognition", "--speech", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--scan-subtitles", "--scan-subtitle", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--extra-media-data", default={})
    parser.add_argument("--extra-playlist-data", default={})

    parser.add_argument("--delete-unplayable", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--force", "-f", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)

    parser.add_argument("database", nargs="?")
    if action == SC.fsadd:
        parser.add_argument("paths", nargs="+")
    args = parser.parse_args()

    if args.db:
        args.database = args.db

    if not args.database:
        if args.profile == DBType.audio:
            args.database = "audio.db"
        elif args.profile == DBType.filesystem:
            args.database = "fs.db"
        elif args.profile == DBType.video:
            args.database = "video.db"
        elif args.profile == DBType.text:
            args.database = "text.db"
        elif args.profile == DBType.image:
            args.database = "image.db"
        else:
            raise Exception(f"fs_extract for profile {args.profile} not implemented")

    if args.db:
        args.database = args.db
    Path(args.database).touch()
    args.db = db.connect(args)
    if hasattr(args, "paths"):
        args.paths = utils.conform(args.paths)
    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def extractor(args, paths) -> None:
    new_files = 0
    for path in paths:
        new_files += scan_path(args, path)

    if not args.db["media"].detect_fts() or new_files > 100000:
        db.optimize(args)


def fs_add(args=None) -> None:
    if args:
        sys.argv[1:] = args

    args = parse_args(
        SC.fsadd,
        """library fsadd [--audio | --video | --image |  --text | --filesystem] -c CATEGORY [database] paths ...

    The default database type is video:
        library fsadd ./tv/
        library fsadd --video ./tv/  # equivalent

    This will create audio.db in the current directory:
        library fsadd --audio ./music/

    This will create image.db in the current directory:
        library fsadd --image ./photos/

    This will create text.db in the current directory:
        library fsadd --text ./documents_and_books/

    Create text database and scan with OCR and speech-recognition:
        library fsadd --text --ocr --speech-recognition ./receipts_and_messages/

    Create video database and read internal/external subtitle files for use in search:
        library fsadd --scan-subtitles ./tv/

    The database location must be specified to reference more than one path:
        library fsadd --audio podcasts.db ./podcasts/ ./another/folder/
    """,
    )

    extractor(args, args.paths)


def fs_update(args=None) -> None:
    if args:
        sys.argv[1:] = args

    args = parse_args(
        SC.fsupdate,
        """library fsupdate database

    Update each path previously saved:

        library fsupdate database
    """,
    )

    playlists = list(
        args.db.query(
            """
            SELECT *
            FROM playlists
            WHERE ie_key = 'Local'
            ORDER BY
                length(path)-length(REPLACE(path, '/', '')) desc
                , path
            """
        )
    )

    for playlist in playlists:
        args_env = args if playlist["config"] is None else argparse.Namespace(**{**playlist["config"], **args.__dict__})
        extractor(args_env, [playlist["path"]])
